import inspect
import sys

from ansys.mapdl.core import LOG
from ansys.mapdl.core.misc import get_active_branch_name


def linkcode_resolve(domain, info, edit=False):
    """Determine the URL corresponding to a Python object.

    Parameters
    ----------
    domain : str
        Only useful when 'py'.

    info : dict
        With keys "module" and "fullname".

    edit : bool, default=False
        Jump right to the edit page.

    Returns
    -------
    url : str
        The code URL.

    Notes
    -----
    This function is used by the extension "sphinx.ext.linkcode" to create the "[Source]"
    button whose link is edited in this function.

    Adapted from mne (mne/utils/docs.py), which was adapted from SciPy (doc/source/conf.py).
    """
    import ansys.mapdl.core as pymapdl
    from ansys.mapdl.core import _version  # noqa: F401
    from ansys.mapdl.core import common_grpc  # noqa: F401
    from ansys.mapdl.core import convert  # noqa: F401
    from ansys.mapdl.core import docs  # noqa: F401
    from ansys.mapdl.core import logging  # noqa: F401
    from ansys.mapdl.core import mapdl  # noqa: F401
    from ansys.mapdl.core import mapdl_console  # noqa: F401
    from ansys.mapdl.core import mapdl_types  # noqa: F401
    from ansys.mapdl.core import mesh_grpc  # noqa: F401
    from ansys.mapdl.core import misc  # noqa: F401
    from ansys.mapdl.core import plotting  # noqa: F401
    from ansys.mapdl.core import post  # noqa: F401
    from ansys.mapdl.core import solution  # noqa: F401
    from ansys.mapdl.core import theme  # noqa: F401
    from ansys.mapdl.core.database import database  # noqa: F401
    from ansys.mapdl.core.database import elems  # noqa: F401
    from ansys.mapdl.core.database import nodes  # noqa: F401

    try:
        from ansys.mapdl.core import mapdl_corba  # noqa: F401
    except ImportError:
        pass
    from ansys.math.core import math  # noqa: F401

    from ansys.mapdl.core import commands  # noqa: F401
    from ansys.mapdl.core import errors  # noqa: F401
    from ansys.mapdl.core import krylov  # noqa: F401
    from ansys.mapdl.core import launcher  # noqa: F401
    from ansys.mapdl.core import licensing  # noqa: F401
    from ansys.mapdl.core import mapdl_geometry  # noqa: F401
    from ansys.mapdl.core import mapdl_grpc  # noqa: F401
    from ansys.mapdl.core import parameters  # noqa: F401
    from ansys.mapdl.core import pool  # noqa: F401
    from ansys.mapdl.core import xpl  # noqa: F401
    from ansys.mapdl.core.examples import downloads  # noqa: F401
    from ansys.mapdl.core.examples import examples  # noqa: F401
    from ansys.mapdl.core.examples import verif_files  # noqa: F401
    from ansys.mapdl.core.inline_functions import core  # noqa: F401
    from ansys.mapdl.core.inline_functions import inline_functions  # noqa: F401

    if domain != "py":  # pragma: no cover
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:  # pragma: no cover
        return None

    obj = submod
    fullname = fullname.replace(modname, "")

    LOG.debug("fullname")
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except AttributeError:
            pass

    # deal with our decorators properly
    while hasattr(obj, "fget"):  # pragma: no cover
        obj = obj.fget

    try:
        fn = inspect.getsourcefile(obj)
    except Exception:  # pragma: no cover
        fn = None
    LOG.debug(f"Filename: {fn}")

    if not fn:  # pragma: no cover
        try:
            fn = inspect.getsourcefile(sys.modules[obj.__module__])
        except Exception:
            return None
        return None

    repo_path = str(pymapdl.__file__)

    if "site-packages" in repo_path:
        src = "site-packages"
    else:  # pragma: no cover
        src = "src"

    repo_path = repo_path[: repo_path.find(src)]

    # Replacing site-packages for source
    fn = fn.replace(repo_path, "").replace("site-packages", "src")

    try:
        source, lineno = inspect.getsourcelines(obj)
    except Exception:  # pragma: no cover
        lineno = None

    if lineno and not edit:
        linespec = f"#L{lineno}-L{lineno + len(source) - 1}"
    else:
        linespec = ""

    kind = get_active_branch_name()
    blob_or_edit = "edit" if edit else "blob"

    url = f"http://github.com/ansys/pymapdl/{blob_or_edit}/{kind}/{fn}{linespec}"
    url = url.replace("\\", "/")  # For windows case

    return url



"""Version of ansys-mapdl-core module.

On the ``main`` branch, use 'dev0' to denote a development version.
For example:

# major, minor, patch
version_info = 0, 58, 'dev0'

"""

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:  # pragma: no cover
    import importlib_metadata

# Read from the pyproject.toml
# major, minor, patch
__version__ = importlib_metadata.version("ansys-mapdl-core")

# In descending order
SUPPORTED_ANSYS_VERSIONS = {
    232: "2023R2",
    231: "2023R1",
    222: "2022R2",
    221: "2022R1",
    212: "2021R2",
    211: "2021R1",
    202: "2020R2",
    201: "2020R1",
    195: "19.5",
    194: "19.4",
    193: "19.3",
    192: "19.2",
    191: "19.1",
}



"""The mapdl database module, allowing the access to the MAPDL database from Python."""

from .database import MINIMUM_MAPDL_VERSION, DBDef, check_mapdl_db_is_alive
from .database import MapdlDb  # noqa: F401



"""Store parameters for a PyMAPDL-specific theme for pyvista"""
from matplotlib.colors import ListedColormap
import numpy as np

from ansys.mapdl.core import _HAS_PYVISTA

if _HAS_PYVISTA:
    from pyvista.plotting.colors import get_cycler

    try:
        from pyvista.plotting.themes import Theme

    except ImportError:
        from pyvista import __version__ as pyvista_version

        if "0.40" in pyvista_version:
            from pyvista.themes import Theme
        else:  # older versions
            from pyvista.themes import DefaultTheme as Theme

    base_class = Theme

else:  # pragma: no cover

    class myEmptyClass:
        pass

    base_class = myEmptyClass


MAPDL_colorbar = (
    np.array(
        [
            [0, 0, 255],
            [0, 178, 255],
            [0, 255, 255],
            [0, 255, 178],
            [0, 255, 0],
            [178, 255, 0],
            [255, 255, 0],
            [255, 178, 0],
            [255, 0, 0],
        ],
        dtype=float,
    )
    / 255
)

PyMAPDL_cmap: ListedColormap = ListedColormap(MAPDL_colorbar, name="PyMAPDL", N=255)


def get_ansys_colors(N=9):
    return np.array([PyMAPDL_cmap(i) for i in range(N)])


class MapdlTheme(base_class):
    """PyMAPDL-specific theme for pyvista.

    Theme includes the following defaults:

    - ``'jet'`` (rainbow) colormap
    - Interactive plots
    - ``'Courier'`` for the font family
    - ``'PyMAPDL'`` as the plot title

    Examples
    --------
    Create a custom theme with unique parameters from the base MapdlTheme.

    >>> from ansys.mapdl import core as pymapdl
    >>> my_theme = pymapdl.MapdlTheme()
    >>> my_theme.background = 'white'
    >>> my_theme.cmap = 'jet'  # colormap
    >>> my_theme.axes.show = False
    >>> my_theme.show_scalar_bar = False

    Apply this theme to element plotting.

    >>> mapdl.eplot(theme=theme)

    Apply this theme to area plotting.

    >>> mapdl.aplot(theme=theme)

    """

    def __init__(self):
        """Initialize the theme."""
        super().__init__()

        self.name = "PyMAPDL-Theme"
        self.title = "PyMAPDL"
        self.background = "paraview"
        self.interactive = True

        self.cmap = PyMAPDL_cmap

        self.font.size = 18
        self.font.title_size = 18
        self.font.label_size = 18
        self.font.color = "black"
        self.show_edges = False
        self.color = "lightblue"
        self.outline_color = "black"
        self.edge_color = "black"

        self.axes.x_color = "tomato"
        self.axes.y_color = "seagreen"
        self.axes.z_color = "blue"

        self.color_cycler = get_cycler(MAPDL_colorbar.tolist())
        self.render_points_as_spheres = True


def _apply_default_theme():
    import pyvista as pv

    pv.global_theme = MapdlTheme()



"""Functions to download sample datasets from the pyansys data repository.
"""
from functools import wraps
import os
import shutil
import urllib.request
import zipfile

import requests

from ansys.mapdl import core as pymapdl


def check_directory_exist(directory):
    # Wrapping LISTING FUNCTIONS.
    def wrap_function(func):
        @wraps(func)
        def inner_wrapper(*args, **kwargs):
            # Check if folder exists
            if not os.path.exists(directory):
                os.makedirs(directory)

            return func(*args, **kwargs)

        return inner_wrapper

    return wrap_function


def get_ext(filename):
    """Extract the extension of the filename"""
    ext = os.path.splitext(filename)[1].lower()
    return ext


def delete_downloads():
    """Delete all downloaded examples to free space or update the files"""
    if os.path.exists(pymapdl.EXAMPLES_PATH):
        shutil.rmtree(pymapdl.EXAMPLES_PATH)
    return True


@check_directory_exist(pymapdl.EXAMPLES_PATH)
def _decompress(filename):
    zip_ref = zipfile.ZipFile(filename, "r")
    zip_ref.extractall(pymapdl.EXAMPLES_PATH)
    return zip_ref.close()


def _get_file_url(filename, directory=None):
    if directory:
        return (
            f"https://github.com/ansys/example-data/raw/master/{directory}/{filename}"
        )
    return f"https://github.com/ansys/example-data/raw/master/{filename}"


def _check_url_exist(url):
    response = requests.get(url)
    if response.status_code == 200:
        return [True]
    else:
        return [False]


@check_directory_exist(pymapdl.EXAMPLES_PATH)
def _retrieve_file(url, filename, _test=False):
    # scape test
    if pymapdl.RUNNING_TESTS:
        return _check_url_exist(url)

    # First check if file has already been downloaded
    local_path = os.path.join(pymapdl.EXAMPLES_PATH, os.path.basename(filename))
    local_path_no_zip = local_path.replace(".zip", "")
    if os.path.isfile(local_path_no_zip) or os.path.isdir(local_path_no_zip):
        return local_path_no_zip, None

    # Perform download
    saved_file, resp = urllib.request.urlretrieve(url)
    shutil.move(saved_file, local_path)
    if get_ext(local_path) in [".zip"]:
        _decompress(local_path)
        local_path = local_path[:-4]
    return local_path, resp


def _download_file(filename, directory=None, _test=False):
    url = _get_file_url(filename, directory)
    try:
        return _retrieve_file(url, filename, _test)
    except Exception as e:  # Genering exception
        raise RuntimeError(
            "For the reason mentioned below, retrieving the file from internet failed.\n"
            "You can download this file from:\n"
            f"{url}\n"
            "\n"
            "The reported error message is:\n"
            f"{str(e)}"
        )


def download_bracket():
    """Download an IGS bracket.

    Examples
    --------
    >>> from ansys.mapdl.core import examples
    >>> filename = examples.download_bracket()
    >>> filename
    '/home/user/.local/share/ansys_mapdl_core/examples/bracket.iges'

    """
    return _download_file("bracket.iges", "geometry")[0]


def download_tech_demo_data(example, filename):
    """Download Tech Demos external data."""
    example = "tech_demos/" + example
    return _download_file(filename=filename, directory=example)[0]


def download_vtk_rotor():
    """Download rotor vtk file."""
    return _download_file("rotor.vtk", "geometry")[0]


def _download_rotor_tech_demo_plot():
    """Download the rotor surface VTK file."""
    return _download_file("rotor2.vtk", "geometry")[0]


def download_example_data(filename, directory=None):
    return _download_file(filename, directory=directory)[0]


def download_manifold_example_data() -> dict:
    """Download the manifold example data and return the
    download paths into a dictionary domain id->path.
    Examples files are downloaded to a persistent cache to avoid
    re-downloading the same file twice.
    Returns
    -------
    dict[str:str]
        Path to the example files.
    Examples
    --------
    Download the manifold geometry, ans file and return the path of the file
    >>> from ansys.mapdl.core.examples import download_manifold_example_data
    >>> paths = download_manifold_example_data()
    >>> paths
    {geometry: 'C:\\Users\\user\\AppData\\Local\\ansys_mapdl_core\\ansys_mapdl_core\\examples\\manifold_geometry.anf',
     mapping_data: 'C:\\Users\\user\\AppData\\Local\\ansys_mapdl_core\\ansys_mapdl_core\\examples\\manifold_cht-final_temp.csv'}
    """

    files_dir = "pymapdl/manifold"
    return {
        "geometry": _download_file(
            filename="manifold_geometry.anf", directory=files_dir
        )[0],
        "mapping_data": _download_file(
            filename="manifold_cht-final_temp.csv", directory=files_dir
        )[0],
    }


def download_cfx_mapping_example_data() -> dict:
    """Download the CFX mapping data and return the
    download paths into a dictionary domain id->path.
    Examples files are downloaded to a persistent cache to avoid
    re-downloading the same file twice.
    Returns
    -------
    dict[str:str]
        Path to the example files.
    Examples
    --------
    >>> from ansys.mapdl.core.examples import download_cfx_mapping_example_data
    >>> paths = download_cfx_mapping_example_data()
    >>> paths
    {data: 'C:\\Users\\user\\AppData\\Local\\ansys_mapdl_core\\ansys_mapdl_core\\examples\\11_blades_mode_1_ND_0.csv',
     model: 'C:\\Users\\user\\AppData\\Local\\ansys_mapdl_core\\ansys_mapdl_core\\examples\\ExampleMapping.db'}
    """

    files_dir = "pymapdl/cfx_mapping"
    return {
        "data": _download_file(
            filename="11_blades_mode_1_ND_0.csv", directory=files_dir
        )[0],
        "model": _download_file(filename="ExampleMapping.db", directory=files_dir)[0],
    }



# Importing logging
import logging
import os
import sys
from warnings import warn

import platformdirs

# Setup data directory
USER_DATA_PATH = platformdirs.user_data_dir(
    appname="ansys_mapdl_core", appauthor="Ansys"
)
if not os.path.exists(USER_DATA_PATH):  # pragma: no cover
    os.makedirs(USER_DATA_PATH)

DEPRECATING_MINIMUM_PYTHON_VERSION = True
MINIMUM_PYTHON_VERSION = (3, 8)

first_time_file = os.path.join(USER_DATA_PATH, ".firstime")
if not os.path.exists(first_time_file):  # pragma: no cover
    py_ver = f"{sys.version_info[0]}.{sys.version_info[1]}"
    py_ver_min = f"{MINIMUM_PYTHON_VERSION[0]}.{MINIMUM_PYTHON_VERSION[1]}"

    if (
        sys.version_info[1] == MINIMUM_PYTHON_VERSION[1]
        and DEPRECATING_MINIMUM_PYTHON_VERSION
    ):
        warn(
            f"Support for Python {py_ver} will be dropped in the next minor " "release."
        )

    if sys.version_info[1] <= MINIMUM_PYTHON_VERSION[1]:
        warn(
            f"Python {py_ver} is not being tested or officially supported. "
            "It is recommended you use a newer version of Python. "
            f"The mininimum supported and tested version is {py_ver_min}.\n\n"
            "**This warning is shown only the first time you run PyMAPDL.**\n"
        )

    with open(first_time_file, "w") as fid:
        fid.write("")

EXAMPLES_PATH = os.path.join(USER_DATA_PATH, "examples")

from ansys.mapdl.core.logging import Logger

LOG = Logger(level=logging.ERROR, to_file=False, to_stdout=True)
LOG.debug("Loaded logging module as LOG")


BUILDING_GALLERY = False
RUNNING_TESTS = False

if RUNNING_TESTS:  # pragma: no cover
    LOG.debug("Running tests on Pytest")

_LOCAL_PORTS = []


# Per contract with Sphinx-Gallery, this method must be available at top level
try:
    import pyvista

    _HAS_PYVISTA = True
except ModuleNotFoundError:  # pragma: no cover
    LOG.debug("The module 'Pyvista' is not installed.")
    _HAS_PYVISTA = False

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:  # pragma: no cover
    import importlib_metadata

__version__ = importlib_metadata.version(__name__.replace(".", "-"))


from ansys.tools.path.path import (
    change_default_ansys_path,
    find_ansys,
    get_ansys_path,
    get_available_ansys_installations,
    save_ansys_path,
)

from ansys.mapdl.core._version import SUPPORTED_ANSYS_VERSIONS
from ansys.mapdl.core.convert import convert_apdl_block, convert_script
from ansys.mapdl.core.launcher import close_all_local_instances

# override default launcher when on pyansys.com
if "ANSJUPHUB_VER" in os.environ:  # pragma: no cover
    from ansys.mapdl.core.jupyter import launch_mapdl_on_cluster as launch_mapdl
else:
    from ansys.mapdl.core.launcher import launch_mapdl

from ansys.mapdl.core.mapdl_grpc import MapdlGrpc as Mapdl
from ansys.mapdl.core.misc import Information, Report, _check_has_ansys
from ansys.mapdl.core.pool import LocalMapdlPool
from ansys.mapdl.core.theme import MapdlTheme, _apply_default_theme

_HAS_ANSYS = _check_has_ansys()

_apply_default_theme()

BUILDING_GALLERY = False
RUNNING_TESTS = False


VERSION_MAP = {
    (0, 0, 0): "2020R2",
    (0, 3, 0): "2021R1",
    (0, 4, 0): "2021R2",
    (0, 4, 1): "2021R2",
    (0, 5, 0): "2022R1",
    (0, 5, 1): "2022R2",  # as of 21 Mar 2022 unreleased
}
"""A dictionary relating PyMAPDL server versions with the unified install ones."""
